/*
 * rng.hpp
 *
 * This file is part of Knights.
 *
 * Copyright (C) Stephen Thompson, 2006 - 2009.
 * Copyright (C) Kalle Marjola, 1994.
 *
 * Knights is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Knights is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Knights.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * Random number generators.
 *
 * getInt, getFloat return number in the range [a,b).
 *
 * TODO: allow save/restore of internal state
 * TODO: improve the seeding algorithm; it seems a waste to use MT but then only use time(0) as the seed!
 * 
 */

#ifndef RNG_HPP
#define RNG_HPP

#include "boost/random/mersenne_twister.hpp"
#include "boost/thread/mutex.hpp"

#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif

class RNG {
public:
    virtual ~RNG() { }

    virtual void seedFromTime() = 0;
    virtual float getU01() = 0;

    bool getBool(float p = 0.5f) { return getU01() <= p; }
    int getInt(int a, int b) { return a + int(getU01() * (b-a)); }
    float getFloat(float a, float b) { return a + getU01() * (b-a); }
};

// for std::random_shuffle compatibility
class RNG_Wrapper {
public:
    RNG_Wrapper(RNG &r_) : r(r_) { }
    int operator()(int n) { return r.getInt(0, n); }
private:
    RNG &r;
};



// Implementations of RNG

// concrete RNG - using rand/srand
// NOT necessarily thread safe.
class StandardRNG : public RNG {
public:
    StandardRNG() { seedFromTime(); }
    virtual void seedFromTime();
    virtual float getU01();
};

// concrete RNG - Mersenne Twister from boost
// Thread safe.
class MersenneRNG : public RNG {
public:
    MersenneRNG() { seedFromTime(); }
    virtual void seedFromTime();
    virtual float getU01();
private:
    typedef boost::mt19937 gen_type;
    gen_type gen;
    boost::mutex mutex;
};

// global access point
extern MersenneRNG g_rng;

#endif      // RNG_HPP
