/*
 * knights_game.cpp
 *
 * This file is part of Knights.
 *
 * Copyright (C) Stephen Thompson, 2006 - 2009.
 * Copyright (C) Kalle Marjola, 1994.
 *
 * Knights is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Knights is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Knights.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "misc.hpp"

#include "anim.hpp"
#include "graphic.hpp"
#include "knights_config.hpp"
#include "knights_engine.hpp"
#include "knights_game.hpp"
#include "knights_log.hpp"
#include "menu.hpp"
#include "menu_constraints.hpp"
#include "menu_selections.hpp"
#include "overlay.hpp"
#include "protocol.hpp"
#include "server_callbacks.hpp"
#include "server_dungeon_view.hpp"
#include "server_mini_map.hpp"
#include "server_status_display.hpp"
#include "sh_ptr_eq.hpp"
#include "sound.hpp"
#include "user_control.hpp"

#include "boost/scoped_ptr.hpp"
#include "boost/weak_ptr.hpp"
#include "boost/thread/barrier.hpp"
#include "boost/thread/locks.hpp"
#include "boost/thread/thread.hpp"

#include <algorithm>
#include <cctype>

#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif

class GameConnection {
public:
    GameConnection(const std::string &n, const std::string &n2, bool new_obs_flag, int ver)
        : name(n), name2(n2),
          is_ready(false), finished_loading(false), ready_to_end(false), 
          obs_flag(new_obs_flag), house_colour(0),
          client_version(ver),
          observer_num(0), 
          ping_time(0)
    { }
    
    std::string name, name2;  // if name2.empty() then normal connection, else split screen connection.
    
    bool is_ready;    // true=ready to start game, false=want to stay in lobby
    bool finished_loading;   // true=ready to play, false=still loading
    bool ready_to_end;   // true=clicked mouse on winner/loser screen, false=still waiting.
    bool obs_flag;   // true=observer, false=player
    int house_colour;  // Must fit in a ubyte. Set to zero for observers.

    int client_version;
    int observer_num;   // 0 if not an observer, or not set yet. >0 if set.
    int player_num;     // 0..num_players-1, or undefined if the game is not running.
    int ping_time;
    
    std::vector<unsigned char> output_data;    

    std::vector<const UserControl*> control_queue[2];
};

typedef std::vector<boost::shared_ptr<GameConnection> > game_conn_vector;

class KnightsGameImpl {
public:
    boost::shared_ptr<const KnightsConfig> knights_config;
    boost::shared_ptr<Coercri::Timer> timer;
    bool allow_split_screen;

    std::vector<const UserControl*> controls;
    
    MenuSelections menu_selections;
    std::string quest_description;
    game_conn_vector connections;

    bool game_over;
    bool pause_mode;
    
    // methods of KnightsGame should (usually) lock this mutex.
    // the update thread will also lock this mutex when it is making changes to the KnightsGameImpl or GameConnection structures.
    boost::mutex my_mutex;

    boost::thread update_thread;

    KnightsLog *knights_log;
    std::string game_name;

    std::vector<int> delete_observer_nums;
    std::vector<int> players_to_eliminate;

    // used during game startup
    boost::scoped_ptr<boost::barrier> barrier;
    std::string startup_err_msg;
};

namespace {

    int GetSelection(const MenuSelections &msel, const std::string &key)
    {
        std::map<std::string, MenuSelections::Sel>::const_iterator it = msel.selections.find(key);
        if (it == msel.selections.end()) return -1;
        else return it->second.value;
    }
    
    std::string GenerateQuestDescription(const MenuSelections &msel, const KnightsConfig &knights_config)
    {
        const int quest = GetSelection(msel, "quest");
        if (quest < 0) {
            return "";
        } else if (quest > 0) {
            return knights_config.getQuestDescription(quest);
        } else {
            std::string result = "Custom Quest\n\n";

            const int mission = GetSelection(msel, "mission");
            const int exit = GetSelection(msel, "exit");
            const int num_gems = GetSelection(msel, "num_gems");
            const int gems_needed = GetSelection(msel, "gems_needed");

            if (mission == 0) {
                result += "You must secure all entry points to prevent your opponent entering the dungeon";
            } else if (mission == 4) {
                result += "You must strike the book with the wand in the special pentagram";
            } else {
                if (mission == 1) {
                    result += "Your mission is to ";
                } else {
                    result += "You must retrieve the ";
                    if (mission == 2) result += "book ";
                    else result += "wand ";
                    result += "and ";
                }
                result += "escape via ";
                switch (exit) {
                case 1:
                    result += "your entry point";
                    break;
                case 2:
                    result += "your opponent's entry point";
                    break;
                case 4:
                    result += "the guarded exit";
                    break;
                default:
                    result += "an unknown exit point";
                    break;
                }
            }
            if (gems_needed > 0) {
                result += " with ";
                result += char('0' + gems_needed);
                result += " out of ";
                result += char('0' + num_gems);
                result += " gems";
            }
            result += ".";

            return result;
        }
    }

    void SendMenuSelections(Coercri::OutputByteBuf &buf, const MenuSelections &msel_old, const MenuSelections &msel_new)
    {
        for (std::map<std::string, MenuSelections::Sel>::const_iterator it = msel_new.selections.begin(); 
        it != msel_new.selections.end(); ++it) {
            std::map<std::string, MenuSelections::Sel>::const_iterator it_old 
                = msel_old.selections.find(it->first);
            const bool needs_update = (it_old == msel_old.selections.end() 
                                     || it_old->second.value != it->second.value
                                     || it_old->second.allowed_values != it->second.allowed_values);
            if (needs_update) {
                buf.writeUbyte(SERVER_SET_MENU_SELECTION);
                buf.writeString(it->first);
                buf.writeVarInt(it->second.value);
                buf.writeVarInt(it->second.allowed_values.size());
                for (std::vector<int>::const_iterator av = it->second.allowed_values.begin(); 
                av != it->second.allowed_values.end(); ++av) {
                    buf.writeVarInt(*av);
                }
            }
        }
    }

    void SendQuestDescription(Coercri::OutputByteBuf &buf, const std::string &descr_old, const std::string &descr_new)
    {
        if (descr_new != descr_old) {
            buf.writeUbyte(SERVER_SET_QUEST_DESCRIPTION);
            buf.writeString(descr_new);
        }
    }
    
    void SendJoinGameAccepted(KnightsGameImpl &impl, Coercri::OutputByteBuf &buf,
                              bool my_obs_flag, int my_house_colour)
    {
        // Send JOIN_GAME_ACCEPTED
        buf.writeUbyte(SERVER_JOIN_GAME_ACCEPTED);
        {
            std::vector<const Graphic*> graphics;
            impl.knights_config->getGraphics(graphics);
            buf.writeVarInt(graphics.size());
            for (int i = 0; i < graphics.size(); ++i) {
                ASSERT(graphics[i]->getID() == i+1);
                graphics[i]->serialize(buf);
            }
        }
        {
            std::vector<const Anim*> anims;
            impl.knights_config->getAnims(anims);
            buf.writeVarInt(anims.size());
            for (int i = 0; i < anims.size(); ++i) {
                ASSERT(anims[i]->getID() == i+1);
                anims[i]->serialize(buf);
            }
        }
        {
            std::vector<const Overlay*> overlays;
            impl.knights_config->getOverlays(overlays);
            buf.writeVarInt(overlays.size());
            for (int i = 0; i < overlays.size(); ++i) {
                ASSERT(overlays[i]->getID() == i+1);
                overlays[i]->serialize(buf);
            }
        }
        {
            std::vector<const Sound*> sounds;
            impl.knights_config->getSounds(sounds);
            buf.writeVarInt(sounds.size());
            for (int i = 0; i < sounds.size(); ++i) {
                ASSERT(sounds[i]->getID() == i+1);
                sounds[i]->serialize(buf);
            }
        }
        {
            std::vector<const UserControl*> controls;
            impl.knights_config->getStandardControls(controls);
            const int num_std_ctrls = controls.size();
            buf.writeVarInt(num_std_ctrls);
            for (int i = 0; i < num_std_ctrls; ++i) {
                ASSERT(controls[i]->getID() == i+1);
                controls[i]->serialize(buf);
            }
            impl.knights_config->getOtherControls(controls);
            buf.writeVarInt(controls.size());
            for (int i = 0; i < controls.size(); ++i) {
                ASSERT(controls[i]->getID() == i+1+num_std_ctrls);
                controls[i]->serialize(buf);
            }
        }
        impl.knights_config->getMenu().serialize(buf);
        buf.writeVarInt(impl.knights_config->getApproachOffset());

        // tell him whether he is an observer or a player, and the house colour
        buf.writeUbyte(my_obs_flag ? 1 : 0);
        buf.writeUbyte(my_house_colour);

        // tell him how many players there are, and their names
        int np = 0;
        for (game_conn_vector::const_iterator it = impl.connections.begin(); it != impl.connections.end(); ++it) {
            if (!(*it)->obs_flag) {
                ++np;
                if (!(*it)->name2.empty()) {
                    if (impl.connections.size() > 1) throw ProtocolError("split screen not supported for network games!");
                    ++np;
                }
            }
        }

        buf.writeVarInt(np);

        for (game_conn_vector::const_iterator it = impl.connections.begin(); it != impl.connections.end(); ++it) {
            if (!(*it)->obs_flag) {
                buf.writeString((*it)->name);
                buf.writeUbyte((*it)->is_ready ? 1 : 0);
                buf.writeUbyte((*it)->house_colour);
                if (!(*it)->name2.empty()) {
                    buf.writeString((*it)->name2);
                    buf.writeUbyte((*it)->is_ready ? 1 : 0);
                    buf.writeUbyte((*it)->house_colour + 1);
                }
            }
        }

        // tell him who the observers are
        // std::max necessary because connections.size()==1 but np==2 in split screen mode.
        buf.writeVarInt(std::max(0, int(impl.connections.size()) - np));
        for (game_conn_vector::const_iterator it = impl.connections.begin(); it != impl.connections.end(); ++it) {
            if ((*it)->obs_flag) {
                ASSERT((*it)->name2.empty());
                buf.writeString((*it)->name);
            }
        }
        
        // Send the current menu selections
        SendMenuSelections(buf, MenuSelections(), impl.menu_selections);
        SendQuestDescription(buf, "", impl.quest_description);

        // Send the available knight house colours
        std::vector<Coercri::Color> hse_cols;
        impl.knights_config->getHouseColours(hse_cols);
        buf.writeUbyte(SERVER_SET_AVAILABLE_HOUSE_COLOURS);
        buf.writeUbyte(hse_cols.size());
        for (int i = 0; i < hse_cols.size(); ++i) {
            buf.writeUbyte(hse_cols[i].r);
            buf.writeUbyte(hse_cols[i].g);
            buf.writeUbyte(hse_cols[i].b);
        }
    }

    int CountPlayers(const KnightsGameImpl &kg)
    {
        int nplayers = 0;
        for (game_conn_vector::const_iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
            if (!(*it)->obs_flag) {
                ++nplayers;
                if (!(*it)->name2.empty()) ++nplayers;
            }
        }
        return nplayers;
    }
    
    // returns true if something was changed.
    bool SetMenuSelection(KnightsGameImpl &impl, const std::string &key, int value)
    {
        // If it's an invalid key then ignore it
        bool do_update = true;
        if (impl.menu_selections.selections.find(key) == impl.menu_selections.selections.end()) do_update = false;
        
        // Save the old settings
        MenuSelections msel_old = impl.menu_selections;
        std::string quest_descr_old = impl.quest_description;

        if (do_update) {
            // If they've changed a non-quest item then change the quest to CUSTOM
            if (key != "quest" && impl.menu_selections.getValue(key) != value) {
                impl.menu_selections.setValue("quest", 0);
            }
            
            // Set the new setting
            impl.menu_selections.setValue(key, value);
        }

        // Count the number of players, this is needed for the constraints
        const int nplayers = CountPlayers(impl);
        
        // Apply menu constraints
        impl.knights_config->getMenuConstraints().apply(impl.knights_config->getMenu(), impl.menu_selections, nplayers);

        // Update the quest description
        impl.quest_description = GenerateQuestDescription(impl.menu_selections, *impl.knights_config);
        
        // Broadcast new menu selections to all clients.
        for (game_conn_vector::iterator it = impl.connections.begin(); it != impl.connections.end(); ++it) {
            Coercri::OutputByteBuf buf((*it)->output_data);
            SendMenuSelections(buf, msel_old, impl.menu_selections);
            SendQuestDescription(buf, quest_descr_old, impl.quest_description);
        }

        // check whether anything changed (for return value)
        bool is_equal = true;
        for (int i = 0; i < impl.knights_config->getMenu().getNumItems(); ++i) {
            const MenuItem &item = impl.knights_config->getMenu().getItem(i);
            const std::map<std::string, MenuSelections::Sel>::const_iterator 
                it_old = msel_old.selections.find(item.getKey()),
                it_new = impl.menu_selections.selections.find(item.getKey());
            is_equal = it_old != msel_old.selections.end()
                && it_new != impl.menu_selections.selections.end()
                && it_old->second.value == it_new->second.value;
            if (!is_equal) break;
        }
        return !is_equal;
    }

    void ResetMenuSelections(KnightsGameImpl &impl)
    {
        // set all menu selections to zero
        for (int i = 0; i < impl.knights_config->getMenu().getNumItems(); ++i) {
            const std::string & key = impl.knights_config->getMenu().getItem(i).getKey();
            impl.menu_selections.setValue(key, 0);
        }

        // now explicitly set "quest" to 1, this will also ensure that menu constraints get set correctly.
        SetMenuSelection(impl, "quest", 1);

        // Update the quest description also
        impl.quest_description = GenerateQuestDescription(impl.menu_selections, *impl.knights_config);
    }
    
    class UpdateThread {
    public:
        explicit UpdateThread(KnightsGameImpl &kg_, boost::shared_ptr<Coercri::Timer> timer_)
            : kg(kg_),
              timer(timer_),
              nplayers(0),
              game_over_sent(false),
              time_to_player_list_update(0),
              prev_total_skulls(0)
        { }

        void operator()()
        {
            try {
            
                // Initialize the game.
                // NOTE: Main thread is waiting on the barrier at this point, so safe to access data from kg

                std::vector<int> hse_cols;
                std::vector<std::string> player_names;
                for (game_conn_vector::const_iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                    if (!(*it)->obs_flag) {
                        hse_cols.push_back((*it)->house_colour);
                        player_names.push_back((*it)->name);
                        if (!(*it)->name2.empty()) {
                            hse_cols.push_back((*it)->house_colour + 1);
                            player_names.push_back((*it)->name2);
                        }
                    }
                }

                if (player_names.size() == 2 && kg.connections.size() == 1) {
                    // Split screen game.
                    // Disable the player names in this case.
                    for (int i = 0; i < player_names.size(); ++i) player_names[i].clear();
                }

                nplayers = player_names.size();
                
                try {
                    engine.reset(new KnightsEngine(kg.knights_config, kg.menu_selections, hse_cols, player_names));
                } catch (const std::exception &e) {
                    kg.startup_err_msg = e.what();
                    if (kg.startup_err_msg.empty()) kg.startup_err_msg = "Unknown error";
                    kg.barrier->wait();
                    return;
                }
                
                callbacks.reset(new ServerCallbacks(hse_cols.size()));
                
                // The game has started successfully so signal the main thread to continue
                kg.barrier->wait();
                
                // Wait until all players have set the 'finished_loading' flag
                while (1) {
                    {
                        boost::lock_guard<boost::mutex> lock(kg.my_mutex);
                        bool all_loaded = true;
                        for (game_conn_vector::const_iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                            if (!(*it)->obs_flag && !(*it)->finished_loading) {
                                all_loaded = false;
                                break;
                            }
                        }
                        if (all_loaded) break;
                    }
                    boost::this_thread::sleep(boost::posix_time::milliseconds(100));
                }
                
                // Go into game loop. NOTE: This will run forever until the main thread interrupts us
                // (or until an exception occurs).
                int last_time = timer->getMsec();
                while (1) {
                    // work out how long since the last update
                    const int new_time = timer->getMsec();
                    const int time_delta = new_time - last_time;

                    // only run an update if delta is min_time_delta or more.
                    // if > max_time_delta then only run an update of max_time_delta, and 'drop'
                    // the rest of the time (ie run more slowly than real-time).
                    const int min_time_delta = 50;
                    const int max_time_delta = 200;
                    if (time_delta > min_time_delta) {
                        update(std::min(time_delta, max_time_delta));
                        last_time += time_delta;
                    }

                    // work out how long until the next update (taking
                    // into account how long the update itself took).
                    const int time_since_update = timer->getMsec() - last_time;
                    const int time_to_wait = std::max(0, min_time_delta - time_since_update) + 1;
                    boost::this_thread::sleep(boost::posix_time::milliseconds(time_to_wait));
                }
                
            } catch (boost::thread_interrupted &) {
                // Allow this to go through. The code that interrupted us knows what it's doing...
                
            } catch (std::exception &e) {
                sendError(e.what());
            } catch (...) {
                sendError("Unknown error in update thread");
            }

            // Before we go, delete the KnightsEngine. This will make sure that the Mediator
            // is still around while the KnightsEngine destructor runs.
            try {
                engine.reset();
            } catch (...) {
                sendError("Error during KnightsEngine shutdown");
            }
        }

        void sendError(const char * msg) throw()
        {
            try {
                // send error to all players connected to this game.
                boost::lock_guard<boost::mutex> lock(kg.my_mutex);
                for (game_conn_vector::const_iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                    Coercri::OutputByteBuf buf((*it)->output_data);
                    buf.writeUbyte(SERVER_ERROR);
                    buf.writeString(msg);
                }
                // log the error as well.
                if (kg.knights_log) {
                    kg.knights_log->logMessage(kg.game_name + "\tError in update thread\t" + msg);
                }
            } catch (...) {
                // Disregard any exceptions here, as we don't want them to propagate up into the
                // boost::thread library internals (this would terminate the entire server if it happened).
            }
        }
        
        void update(int time_delta)
        {
            // read controls (and see if paused)
            {
                boost::lock_guard<boost::mutex> lock(kg.my_mutex);

                // Delete any "dead" observers.
                for (std::vector<int>::iterator it = kg.delete_observer_nums.begin(); it != kg.delete_observer_nums.end(); ++it) {
                    callbacks->rmObserverNum(*it);
                }
                kg.delete_observer_nums.clear();
                
                // Delete any disconnected players
                for (std::vector<int>::iterator it = kg.players_to_eliminate.begin(); it != kg.players_to_eliminate.end(); ++it) {
                    engine->eliminatePlayer(*it);
                }
                kg.players_to_eliminate.clear();

                // we assume that allow_split_screen means this is a local game
                // (and therefore pause should be allowed)
                if (kg.pause_mode && kg.allow_split_screen) {
                    return;  // ignore the update
                }

                for (game_conn_vector::const_iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                    if (!(*it)->obs_flag) {
                        for (int p = 0; p < ((*it)->name2.empty() ? 1 : 2); ++p) {

                            const UserControl *final_ctrl = 0;

                            for (std::vector<const UserControl*>::const_iterator c = (*it)->control_queue[p].begin();
                            c != (*it)->control_queue[p].end(); ++c) {
                                final_ctrl = *c;
                                engine->setControl((*it)->player_num + p, *c);
                            }
                            (*it)->control_queue[p].clear();

                            // if the final control is continuous then re-insert it into the queue for next time.
                            if (final_ctrl && final_ctrl->isContinuous()) {
                                (*it)->control_queue[p].push_back(final_ctrl);
                            }
                        }
                    } else {
                        // He is an observer. If he has not been allocated an observer_num
                        // yet then give him one, and send him the startup commands. (but only if
                        // he has finished loading.)
                        if ((*it)->observer_num == 0 && (*it)->finished_loading) {
                            std::vector<unsigned char> & buf = (*it)->output_data;

                            for (int p = 0; p < nplayers; ++p) {
                                buf.push_back(SERVER_SWITCH_PLAYER);
                                buf.push_back(p);

                                // slight hack -- create 'temporary' dungeonview/minimap objects.
                                // This is the simplest way to get the data sent only to this new observer
                                // and no-one else.
                                ServerDungeonView dungeon_view(buf);
                                ServerMiniMap mini_map(buf);
                                ServerStatusDisplay status_display(buf);
                                engine->catchUp(p, dungeon_view, mini_map, status_display);
                                mini_map.appendMiniMapCmds(buf);
                                dungeon_view.appendDungeonViewCmds(1000 + p, buf);  // obs_num can be a dummy because we're throwing away the caches anyway
                                // no need to call 'clearCmds' because the dview/mini map are being deleted anyway.
                            }
                            
                            (*it)->observer_num = callbacks->allocObserverNum();
                        }
                    }
                }
            }
            
            // run the update
            engine->update(time_delta, *callbacks);

            // copy the results back to the GameConnection buffers
            {
                boost::lock_guard<boost::mutex> lock(kg.my_mutex);
                for (game_conn_vector::iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                    if ((*it)->obs_flag) {
                        if ((*it)->observer_num > 0) {
                            // Send observer cmds to that player. (We're assuming observation is always allowed at the moment.)
                            callbacks->appendObserverCmds((*it)->observer_num, (*it)->output_data);
                        }
                    } else {
                        const bool split_screen = !(*it)->name2.empty();
                        if (split_screen) {
                            // Select player 0 for split screen mode
                            (*it)->output_data.push_back(SERVER_SWITCH_PLAYER);
                            (*it)->output_data.push_back(0);
                        }

                        const int buf_size_before_p0 = (*it)->output_data.size();
                        callbacks->appendPlayerCmds((*it)->player_num, (*it)->output_data);

                        if (split_screen) {
                            if ((*it)->output_data.size() == buf_size_before_p0) {
                                // can remove the SWITCH_PLAYER 0 cmd
                                (*it)->output_data.pop_back();
                                (*it)->output_data.pop_back();
                            }

                            // Select player 1 for split screen mode
                            (*it)->output_data.push_back(SERVER_SWITCH_PLAYER);
                            (*it)->output_data.push_back(1);

                            // Send player 1's data
                            const int buf_size_before_p1 = (*it)->output_data.size();
                            callbacks->appendPlayerCmds((*it)->player_num + 1, (*it)->output_data);

                            if ((*it)->output_data.size() == buf_size_before_p1) {
                                // can remove the SWITCH_PLAYER 1 cmd
                                (*it)->output_data.pop_back();
                                (*it)->output_data.pop_back();
                            }
                        }
                    }
                }
                callbacks->clearCmds();
            }

            // do player list update every N seconds, or if someone has died (total skulls has changed)
            const int total_skulls = engine->getSkullsPlusKills();
            time_to_player_list_update -= time_delta;
            if (time_to_player_list_update <= 0 || total_skulls != prev_total_skulls) {

                std::vector<PlayerInfo> player_list;
                engine->getPlayerList(player_list);

                boost::lock_guard<boost::mutex> lock(kg.my_mutex);

                // associate a ping time with each player
                // -- Only do this if the timer ran out, not if we are updating following a death,
                // otherwise it looks strange when the pings update just because someone died.
                if (time_to_player_list_update <= 0) {
                    for (int idx = 0; idx < player_list.size(); ++idx) {
                        const int player_num = player_list[idx].player_num;
                        const std::string &player_name = player_list[idx].name;
                        for (game_conn_vector::const_iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                            if ((*it)->player_num == player_num) {
                                pings[player_name] = (*it)->ping_time;
                            }
                        }
                    }
                }

                // now send out the updates
                for (game_conn_vector::const_iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                    Coercri::OutputByteBuf buf((*it)->output_data);
                    buf.writeUbyte(SERVER_PLAYER_LIST);
                    buf.writeVarInt(player_list.size());
                    for (int idx = 0; idx < player_list.size(); ++idx) {
                        buf.writeString(player_list[idx].name);
                        buf.writeUbyte(player_list[idx].house_colour.r);
                        buf.writeUbyte(player_list[idx].house_colour.g);
                        buf.writeUbyte(player_list[idx].house_colour.b);
                        buf.writeVarInt(player_list[idx].kills);
                        buf.writeVarInt(player_list[idx].deaths);
                        buf.writeVarInt(pings[player_list[idx].name]);
                    }
                }

                // Reset the update timers etc
                prev_total_skulls = total_skulls;
                if (time_to_player_list_update <= 0) time_to_player_list_update = 3000;  // 3 seconds
            }

            // poll to see if the game is over.
            if (!game_over_sent && callbacks->isGameOver()) {
                boost::lock_guard<boost::mutex> lock(kg.my_mutex);
                kg.game_over = true;
                game_over_sent = true;

                // find names of all players, and the winner
                std::vector<std::string> loser_names;
                std::string winner_name;
                for (game_conn_vector::const_iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                    if (!(*it)->obs_flag) {
                        if ((*it)->player_num == callbacks->getWinnerNum()) {
                            winner_name = (*it)->name;
                        } else {
                            loser_names.push_back((*it)->name);
                        }
                        // note: name2 not supported here currently.
                    }
                }
                
                if (kg.knights_log) {
                    std::string msg;
                    msg = kg.game_name + "\tgame won\t";
                    // the first name printed is the winner, then we list the losers (of which there will be
                    // only one at the moment).
                    msg += "winner=" + winner_name;
                    for (std::vector<string>::const_iterator it = loser_names.begin(); it != loser_names.end(); ++it) {
                        msg += std::string(", loser=") + *it;
                    }
                    kg.knights_log->logMessage(msg);
                }
            }
        }

    private:
        KnightsGameImpl &kg;
        boost::shared_ptr<Coercri::Timer> timer;
        boost::shared_ptr<ServerCallbacks> callbacks;
        boost::shared_ptr<KnightsEngine> engine;
        std::map<std::string, int> pings;
        int nplayers;
        bool game_over_sent;
        int time_to_player_list_update;
        int prev_total_skulls;
    };

    void DoSetReady(KnightsGameImpl &kg, GameConnection &conn, bool ready)
    {
        conn.is_ready = ready;

        // send notification to players
        for (game_conn_vector::iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
            Coercri::OutputByteBuf out((*it)->output_data);
            out.writeUbyte(SERVER_SET_READY);
            out.writeString(conn.name);
            out.writeUbyte(ready ? 1 : 0);
            if (!conn.name2.empty()) {
                out.writeUbyte(SERVER_SET_READY);
                out.writeString(conn.name2);
                out.writeUbyte(ready ? 1 : 0);
            }
        }
    }
    
    struct CmpByName {
        bool operator()(const boost::shared_ptr<GameConnection> &lhs,
                        const boost::shared_ptr<GameConnection> &rhs) const
        {
            return ToUpper(lhs->name) < ToUpper(rhs->name);
        }
        static std::string ToUpper(const std::string &input)
        {
            std::string result;
            for (std::string::const_iterator it = input.begin(); it != input.end(); ++it) {
                result += std::toupper(*it);
            }
            return result;
        }
    };

    void StartGameIfReady(KnightsGameImpl &kg)
    {
        // mutex is locked at this point

        std::vector<std::string> names;
        
        // count how many players are ready to leave the lobby, and how many there are total
        int nready = 0;
        int nplayers = 0;
        for (game_conn_vector::iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
            if (!(*it)->obs_flag) {
                (*it)->player_num = nplayers;  // assign player_nums
                const int ncount = (*it)->name2.empty() ? 1 : 2;
                nplayers += ncount;
                if ((*it)->is_ready) {
                    nready += ncount;
                    names.push_back((*it)->name);
                    if (ncount==2) names.push_back((*it)->name2);
                }
            }
        }

        // if all players are ready, and there are at least two players, then can start the game
        if (nready == nplayers && nplayers >= 2) {
            // Start the update thread.
            kg.barrier.reset(new boost::barrier(2));
            kg.startup_err_msg.clear();
            UpdateThread thr(kg, kg.timer);
            boost::thread new_thread(thr);
            kg.update_thread.swap(new_thread);  // start the sub-thread -- game is now running.

            // Wait for barrier before doing anything else...
            kg.barrier->wait();

            // We're done with the barrier, can get rid of it now.
            kg.barrier.reset();
            
            // If the game failed to initialize then print a message and refuse to start the game.
            // Otherwise, send startup message to all players.
            const std::string err_msg = kg.startup_err_msg;
            if (!err_msg.empty()) {
                for (game_conn_vector::const_iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                    Coercri::OutputByteBuf buf((*it)->output_data);
                    buf.writeUbyte(SERVER_ANNOUNCEMENT);
                    buf.writeString(std::string("Couldn't start game! ") + err_msg);
                }

                // log the error as well
                if (kg.knights_log) {
                    kg.knights_log->logMessage(kg.game_name + "\tError starting game\t" + err_msg);
                }
                
                kg.update_thread.join();

            } else {
                // No errors -- We can start the game now

                // Clear 'finished_loading' for each player; assign player nums; send startGame msg
                for (game_conn_vector::iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
                    (*it)->finished_loading = false;
                    (*it)->ready_to_end = false;
                    
                    unsigned char num_displays;
                    if (!(*it)->name2.empty()) num_displays = 2;  // split screen mode
                    else if ((*it)->obs_flag) num_displays = nplayers;  // observer mode
                    else num_displays = 1;   // normal mode
                    
                    if (!(*it)->obs_flag) {
                        // Player.
                        Coercri::OutputByteBuf buf((*it)->output_data);
                        buf.writeUbyte(SERVER_START_GAME);
                        buf.writeUbyte(num_displays);
                    } else {
                        // Observer.
                        Coercri::OutputByteBuf buf((*it)->output_data);
                        buf.writeUbyte(SERVER_START_GAME_OBS);
                        buf.writeUbyte(num_displays);
                        for (int i = 0; i < num_displays; ++i) {
                            buf.writeString(names[i]);
                        }
                        buf.writeUbyte(0);  // already_started flag (false)
                    }
                    
                    // clear all ready flags when the game starts.
                    (*it)->is_ready = false;
                }

                // Clear 'players_to_eliminate' in case there is anything still in there from the previous game
                kg.players_to_eliminate.clear();
            
                // Log a message if required.
                if (kg.knights_log) {
                    std::ostringstream str;
                    str << kg.game_name << "\tgame started\t";
                    for (std::vector<std::string>::const_iterator it = names.begin(); it != names.end(); ++it) {
                        if (it != names.begin()) str << ", ";
                        str << *it;
                    }
                    
                    for (std::map<std::string, MenuSelections::Sel>::const_iterator it = kg.menu_selections.selections.begin();
                    it != kg.menu_selections.selections.end(); ++it) {
                        // TODO: could try to convert this to a summary message rather than dumping all settings?
                        str << ", " << it->first << "=" << it->second.value;
                    }
                    
                    kg.knights_log->logMessage(str.str());
                }
            }
        }
    }

    void StopGameAndReturnToMenu(KnightsGameImpl &kg)
    {
        // Interrupt the thread and tell all players to return to menu.
        // Called at game over (after all players have clicked mouse to continue)
        // or when somebody presses escape.

        // NOTE: Mutex should be UNLOCKED when calling this routine, otherwise we will get a 
        // deadlock if the update thread is currently waiting for the mutex. (Apparently
        // a thread can't be interrupted while it is waiting for a mutex.)
        
        kg.update_thread.interrupt();
        kg.update_thread.join();
        for (game_conn_vector::iterator it = kg.connections.begin(); it != kg.connections.end(); ++it) {
            (*it)->output_data.push_back(SERVER_GOTO_MENU);
            (*it)->observer_num = 0;
            (*it)->player_num = 0;
        }

        // Take it out of pause mode if necessary
        kg.pause_mode = false;
        
        // Log end of game.
        if (kg.knights_log) {
            kg.knights_log->logMessage(kg.game_name + "\tgame ended");
        }
    }
}

KnightsGame::KnightsGame(boost::shared_ptr<const KnightsConfig> config,
                         boost::shared_ptr<Coercri::Timer> tmr,
                         bool allow_split_screen,
                         KnightsLog *knights_log,
                         const std::string &game_name)
    : pimpl(new KnightsGameImpl)
{
    pimpl->knights_config = config;
    pimpl->timer = tmr;
    pimpl->allow_split_screen = allow_split_screen;
    pimpl->game_over = false;
    pimpl->pause_mode = false;
    
    // initialize all menu settings to 0
    ResetMenuSelections(*pimpl);

    // set up our own controls vector.
    config->getStandardControls(pimpl->controls);
    std::vector<const UserControl*> other_ctrls;
    config->getOtherControls(other_ctrls);
    pimpl->controls.insert(pimpl->controls.end(), other_ctrls.begin(), other_ctrls.end());

    pimpl->knights_log = knights_log;
    pimpl->game_name = game_name;
}

KnightsGame::~KnightsGame()
{
    // If the update thread is still running then we have to kill it
    // here. The thread dtor will not kill the thread for us, instead
    // the thread will just become "detached" and continue running,
    // with disastrous consequences if the KnightsGame object is being
    // destroyed!
    pimpl->update_thread.interrupt();
    pimpl->update_thread.join();
}

int KnightsGame::getNumPlayers() const
{
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    return CountPlayers(*pimpl);
}

int KnightsGame::getNumObservers() const
{
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    int n = 0;
    for (game_conn_vector::const_iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
        if ((*it)->obs_flag) ++n;
    }
    return n;
}

bool KnightsGame::isSplitScreenAllowed() const
{
    // sub thread won't be changing allow_split_screen so don't need to lock
    return pimpl->allow_split_screen;
}

GameStatus KnightsGame::getStatus() const
{
    const int nplayers = getNumPlayers();
    if (nplayers < 2) return GS_WAITING_FOR_PLAYERS;
    if (pimpl->update_thread.joinable()) return GS_RUNNING;
    else return GS_SELECTING_QUEST;
}

bool KnightsGame::getObsFlag(GameConnection &conn) const
{
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    return conn.obs_flag;
}

GameConnection & KnightsGame::newClientConnection(const std::string &client_name, const std::string &client_name_2, int client_version)
{
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);

    // check preconditions. (caller should have checked these already.)
    if (client_name.empty()) throw UnexpectedError("invalid client name");
    for (game_conn_vector::const_iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
        if ((*it)->name == client_name) throw UnexpectedError("Duplicate client name");
        if (!client_name_2.empty() && (*it)->name == client_name_2) throw UnexpectedError("Duplicate client name");
    }
    if (!pimpl->allow_split_screen && !client_name_2.empty()) throw UnexpectedError("Split screen mode not allowed");
    if (!client_name_2.empty() && !pimpl->connections.empty()) throw UnexpectedError("Cannot join in split screen mode while connections exist");

    // If the game is running, they join as an observer, otherwise, they join as a player
    const bool new_obs_flag = pimpl->update_thread.joinable();

    // create the GameConnection
    boost::shared_ptr<GameConnection> conn(new GameConnection(client_name, client_name_2, new_obs_flag, client_version));
    pimpl->connections.push_back(conn);

    // sort connections by name
    // (this makes the player list appear in alphabetical order.)
    std::sort(pimpl->connections.begin(), pimpl->connections.end(), CmpByName());

    // modify the house colour if necessary.
    if (!new_obs_flag) {
        bool col_ok = false;
        while (!col_ok) {
            col_ok = true;
            for (game_conn_vector::const_iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
                if (*it != conn && (*it)->house_colour == conn->house_colour && !(*it)->obs_flag) {
                    col_ok = false;
                    // try the next colour
                    conn->house_colour ++;
                    std::vector<Coercri::Color> hse_cols;
                    pimpl->knights_config->getHouseColours(hse_cols);  // a bit wasteful as we only want the size. never mind.
                    if (conn->house_colour == hse_cols.size() - 1) {
                        // run out of colours so will just have to accept the last one
                        col_ok = true; 
                    }
                    break;
                }
            }
        }
    }

    // Send the SERVER_JOIN_GAME_ACCEPTED message (includes initial configuration messages e.g. menu settings)
    Coercri::OutputByteBuf buf(conn->output_data);
    SendJoinGameAccepted(*pimpl, buf, new_obs_flag, conn->house_colour);

    // Send any necessary SERVER_PLAYER_JOINED_THIS_GAME messages (but not to the player who just joined).
    for (game_conn_vector::iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
        if (*it != conn) {
            Coercri::OutputByteBuf out((*it)->output_data);
            out.writeUbyte(SERVER_PLAYER_JOINED_THIS_GAME);
            out.writeString(client_name);
            out.writeUbyte(new_obs_flag);
            out.writeUbyte(conn->house_colour);
            if (!client_name_2.empty()) {
                out.writeUbyte(SERVER_PLAYER_JOINED_THIS_GAME);
                out.writeString(client_name_2);
                out.writeUbyte(new_obs_flag);
                out.writeUbyte(conn->house_colour);
            }
        }
    }

    // If the new player is an observer and the game is running then
    // send him the START_GAME_OBS msg immediately
    if (new_obs_flag && pimpl->update_thread.joinable()) {
        // find the player names
        std::vector<std::string> player_names;
        for (game_conn_vector::const_iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
            if (!(*it)->obs_flag) {
                player_names.push_back((*it)->name);
            }
        }

        // send the msg
        buf.writeUbyte(SERVER_START_GAME_OBS);
        buf.writeUbyte(player_names.size());
        for (std::vector<std::string>::const_iterator it = player_names.begin(); it != player_names.end(); ++it) {
            buf.writeString(*it);
        }
        buf.writeUbyte(1);  // already_started flag (true)
    }

    if (!new_obs_flag) {
        // May need to update menu settings, since no of players has changed
        SetMenuSelection(*pimpl, "", 0);
    }
    
    return *conn;
}

void KnightsGame::clientLeftGame(GameConnection &conn)
{
    // This is called when a player has left the game (either because
    // KnightsServer sent him a SERVER_LEAVE_GAME msg, or because he
    // disconnected.)

    bool is_player = false;
    
    std::string name, name2;
    int player_num = 0;
    int num_player_connections = 0;

    {
        boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
        
        // Find the client in our connections list. 
        game_conn_vector::iterator where = std::find_if(pimpl->connections.begin(), 
                                                        pimpl->connections.end(), ShPtrEq<GameConnection>(&conn));
        ASSERT(where != pimpl->connections.end());
        
        // Determine whether he is a player or observer. Also save his name(s)
        is_player = !(*where)->obs_flag;
        player_num = (*where)->player_num;
        name = (*where)->name;
        name2 = (*where)->name2;
        if ((*where)->observer_num > 0) {
            pimpl->delete_observer_nums.push_back( (*where)->observer_num );
        }
        
        // Erase the connection from our list. Note 'conn' is invalid from now on
        pimpl->connections.erase(where);

        // find out how many player connections are remaining
        for (game_conn_vector::const_iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
            if (!(*it)->obs_flag) {
                ++num_player_connections;
            }
        }
    }
    
    if (is_player && pimpl->update_thread.joinable()) {
        // The leaving client is one of the players (as opposed to an observer).
        if (num_player_connections == 1) {
            // Only one player connection is left. We should just terminate the game at this point.
            StopGameAndReturnToMenu(*pimpl);
        } else {
            // There are still players left in so eliminate that player and allow the game to continue.
            boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
            pimpl->players_to_eliminate.push_back(player_num);
        }
    }

    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);

    // Send SERVER_PLAYER_LEFT_THIS_GAME message to all remaining players
    for (game_conn_vector::iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
        Coercri::OutputByteBuf buf((*it)->output_data);
        buf.writeUbyte(SERVER_PLAYER_LEFT_THIS_GAME);
        buf.writeString(name);
        buf.writeUbyte((*it)->obs_flag ? 1 : 0);
        if (!name2.empty()) {
            buf.writeUbyte(SERVER_PLAYER_LEFT_THIS_GAME);
            buf.writeString(name2);
            buf.writeUbyte((*it)->obs_flag ? 1 : 0);
        }
    }

    if (pimpl->connections.empty()) {
        // If there are no connections left then reset the menu selections.
        ResetMenuSelections(*pimpl);
    } else if (is_player) {
        // Number of players has changed, may need to update menu constraints.
        SetMenuSelection(*pimpl, "", 0);
    }
}

void KnightsGame::sendChatMessage(GameConnection &conn, const std::string &msg)
{
    // Forward the message to everybody (including the originator)
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    for (game_conn_vector::iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
        Coercri::OutputByteBuf buf((*it)->output_data);
        buf.writeUbyte(SERVER_CHAT);
        buf.writeString(conn.name);
        buf.writeUbyte(conn.obs_flag ? 2 : 1);  // 2 means observer, 1 means player.
        buf.writeString(msg);
    }
}

void KnightsGame::setReady(GameConnection &conn, bool ready)
{
    if (pimpl->update_thread.joinable()) return;  // Game is running.

    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    if (!conn.obs_flag) {
        // Set conn.is_ready and send out notifications
        DoSetReady(*pimpl, conn, ready);

        // Start the game if all players are now ready.
        StartGameIfReady(*pimpl);        
    }
}

void KnightsGame::setHouseColour(GameConnection &conn, int hse_col)
{
    if (pimpl->update_thread.joinable()) return;  // Game is running

    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    if (!conn.obs_flag) {

        conn.house_colour = hse_col;

        // send notification to players
        for (game_conn_vector::iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
            Coercri::OutputByteBuf out((*it)->output_data);
            out.writeUbyte(SERVER_SET_HOUSE_COLOUR);
            // note we don't support house colours in split screen mode at the moment. we assume it's the first player
            // on the connection who is having the house colours set.
            out.writeString(conn.name);
            out.writeUbyte(hse_col);
            out.writeUbyte(it->get() == &conn ? 1 : 0);
        }
    }
}
    
void KnightsGame::finishedLoading(GameConnection &conn)
{
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    conn.finished_loading = true;
}

void KnightsGame::readyToEnd(GameConnection &conn)
{
    if (!pimpl->game_over) return;
    
    bool all_ready = false;
    {
        boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
        if (!conn.obs_flag) {
            conn.ready_to_end = true;
            all_ready = true;
            for (game_conn_vector::const_iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
                if (!(*it)->obs_flag && !(*it)->ready_to_end) {
                    all_ready = false;
                    break;
                }
            }
        }
    }
    
    if (all_ready) {
        StopGameAndReturnToMenu(*pimpl);
    }
}

bool KnightsGame::requestQuit(GameConnection &conn)
{
    if (!pimpl->update_thread.joinable()) return false;  // Game is not running
    
    bool obs_flag;
    {
        boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
        obs_flag = conn.obs_flag;
    }
    if (!obs_flag) {

        if (pimpl->knights_log) {
            pimpl->knights_log->logMessage(pimpl->game_name + "\tquit requested\t" + conn.name);
        }

        // tell all players what happened.
        {
            boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
            for (game_conn_vector::iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
                Coercri::OutputByteBuf buf((*it)->output_data);
                buf.writeUbyte(SERVER_ANNOUNCEMENT);
                buf.writeString(conn.name + " quit the game.");
            }
        }

        // If there are only two players then stop the game, otherwise, go back to the lobby
        if (getNumPlayers() > 2) {
            boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
            pimpl->players_to_eliminate.push_back(conn.player_num);
            return true;
        } else {
            StopGameAndReturnToMenu(*pimpl);
            return false;
        }
    }

    return false;
}

void KnightsGame::setPauseMode(bool pm)
{
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    pimpl->pause_mode = pm;
}

void KnightsGame::setMenuSelection(GameConnection &conn, const std::string &key, int value)
{
    if (pimpl->update_thread.joinable()) return;  // Game is running
    
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    if (conn.obs_flag) return;   // only players can adjust the menu.

    const bool changed = SetMenuSelection(*pimpl, key, value);

    if (changed) {
        // send announcement to all players.
        int x = 0;
        for (; x < pimpl->knights_config->getMenu().getNumItems(); ++x) {
            if (pimpl->knights_config->getMenu().getItem(x).getKey() == key) break;
        }
        std::string str;
        if (x >= 0 && x < pimpl->knights_config->getMenu().getNumItems()) {
            const MenuItem &item = pimpl->knights_config->getMenu().getItem(x);
            str = conn.name + " set \"" + item.getTitleString() + '"';
            std::map<std::string, MenuSelections::Sel>::const_iterator it = pimpl->menu_selections.selections.find(key);
            if (it != pimpl->menu_selections.selections.end()) {
                const std::string val_str = item.getValueString(it->second.value);
                
                bool all_digit = true;
                for (std::string::const_iterator v = val_str.begin(); v != val_str.end(); ++v) {
                    if (!std::isdigit(*v)) {
                        all_digit = false;
                        break;
                    }
                }

                str += " to ";
                if (!all_digit) str += '"';
                str += item.getValueString(it->second.value);
                if (!all_digit) str += '"';
                str += '.';
            }
        }
        for (game_conn_vector::iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
            Coercri::OutputByteBuf buf((*it)->output_data);
            buf.writeUbyte(SERVER_ANNOUNCEMENT);
            buf.writeString(str);
        }
    }
}

void KnightsGame::sendControl(GameConnection &conn, int p, unsigned char control_num)
{
    if (!pimpl->update_thread.joinable()) return; // Game is not running
    
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    if (!conn.obs_flag && p >= 0 && p < (conn.name2.empty() ? 1 : 2)) {
        const UserControl * control = control_num == 0 ? 0 : pimpl->controls.at(control_num - 1);
        conn.control_queue[p].push_back(control);
    }
}

void KnightsGame::setObsFlag(GameConnection &conn, bool new_obs_flag)
{
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);

    if (conn.obs_flag == new_obs_flag) return;  // Nothing to do
    
    // If the game is in progress then don't allow players to change obs-flag setting
    if (pimpl->update_thread.joinable()) {
        Coercri::OutputByteBuf buf(conn.output_data);
        buf.writeUbyte(SERVER_ANNOUNCEMENT);
        buf.writeString("Cannot change observer status at this time. Please wait until the game has finished.");
        return;
    }

    // If he is becoming an observer, and he is currently 'ready to start', then make him non-ready first.
    if (new_obs_flag && conn.is_ready) {
        DoSetReady(*pimpl, conn, false);
    }
    
    // changing obs_flag does not apply to split screen games, so we can ignore name2 and just use name.
    const std::string & my_name = conn.name;

    // If he is a player then assign him a house colour.
    int new_col = 0;
    if (!new_obs_flag) {
        std::vector<Coercri::Color> hse_cols;
        pimpl->knights_config->getHouseColours(hse_cols);

        while (new_col < int(hse_cols.size()) - 1) {
            bool col_taken = false;
            for (game_conn_vector::iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
                if (!(*it)->obs_flag && (*it)->house_colour == new_col) {
                    col_taken = true;
                    break;
                }
            }
            if (col_taken) ++new_col;
            else break;
        }
    }

    // Update our data, then send announcements to all players.
    conn.obs_flag = new_obs_flag;
    conn.house_colour = new_col;
    for (game_conn_vector::iterator it = pimpl->connections.begin(); it != pimpl->connections.end(); ++it) {
        const bool me = (it->get() == &conn);
        Coercri::OutputByteBuf buf((*it)->output_data);

        buf.writeUbyte(SERVER_SET_OBS_FLAG);
        buf.writeString(my_name);
        buf.writeUbyte(new_obs_flag ? 1 : 0);
        buf.writeUbyte(me ? 1 : 0);
        
        if (!new_obs_flag) {
            buf.writeUbyte(SERVER_SET_HOUSE_COLOUR);
            buf.writeString(my_name);
            buf.writeUbyte(new_col);
            buf.writeUbyte(me ? 1 : 0);
        }
    }

    // May need to re-apply max players constraints
    SetMenuSelection(*pimpl, "", 0);
}

void KnightsGame::getOutputData(GameConnection &conn, std::vector<unsigned char> &data)
{
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    std::swap(conn.output_data, data);
    conn.output_data.clear();
}

void KnightsGame::setPingTime(GameConnection &conn, int ping)
{
    boost::lock_guard<boost::mutex> lock(pimpl->my_mutex);
    conn.ping_time = ping;
}
